
assert( LibStub, "LibStub is not loaded." )

local addon = LibStub( "AceAddon-3.0" ):NewAddon( "hcRepairBot", "AceConsole-3.0", "AceEvent-3.0" )

local DEFAULTS = {
  profile = {
		-- True if this addon should be enabled.
		enabled = true,
		-- True if debugging message should be output by this addon.
		verbose = false,
		player = {
			-- True if player-funded repairs should be permitted in any 
			-- situation.
			enabled = true,
			-- The minimum faction standing with the merchant to permit 
			-- player-funded repairs.
			reputation = 0,
			-- The minimum amount of money the player should reserve. 
			reserve = 0,
		},
		guild = {
			-- True if guild-funded repairs should be permitted in any 
			-- situation.
			raid = {
				-- True if guild-funded repairs should be permitted when the 
				-- player is in a guild raid.
				enabled = true,
				-- True if guild-funded repairs should be permitted when the 
				-- player is in any raid.				
				any = false,
			},
			party = {
				-- True if guild-funded repairs should be permitted when the 
				-- player is in a guild party.
				enabled = false,
				-- True if guild-funded repairs should be permitted when the 
				-- player is in any party.
				any = false, 
			},
			solo = {
				-- True if guild-funded repairs should be permitted when the
				-- player is not in a group.
				enabled = false, 
			},
			-- The minimum faction standing with the merchant to permit 
			-- guild-funded repairs.
			reputation = 0,
		},
  },
}

local function _IsAddonDisabled( info )
	return not addon.db.profile.enabled
end

local function _IsPlayerRepairDisabled( info )
	return not addon.db.profile.player.enabled or _IsAddonDisabled( info )
end

local function _IsGuildRepairDisabled( info )
	return not addon.db.profile.guild.enabled or _IsAddonDisabled( info )
end

local function _IsPartyRepairDisabled( info )
	return not addon.db.profile.guild.party.enabled or _IsGuildRepairDisabled( info )
end

local function _IsRaidRepairDisabled( info )
	return not addon.db.profile.guild.raid.enabled or _IsGuildRepairDisabled( info )
end

local OPTIONS_REPUTATION_VALUES = {
	[ 0 ] = "None",
	[ 5 ] = "Friendly",
	[ 6 ] = "Honored",
	[ 7 ] = "Revered",
	[ 8 ] = "Exalted",
}

local OPTIONS = {
	type = "group",
	name = "Repair",
	args = {
		enable = {
			type = "toggle",
			order = 1,
			width = "full",
			name = "Enabled",
			desc = "Enables or disables automatic repairs.",
			descStyle = "inline",
			get = function( info ) return addon.db.profile.enabled end,
			set = function( info, value ) if value then addon:Enable( ) else addon:Disable( ) end end,
			disabled = false,
		},
		verbose = {
			type = "toggle",
			order = 2,
			width = "full",
			name = "Verbose",
			desc = "Enables or disables verbose output messages about repair costs and other details.",
			descStyle = "inline",
			get = function( info ) return addon.db.profile.verbose end,
			set = function( info, value ) addon.db.profile.verbose = value end,
			disabled = _IsAddonDisabled,
		},
		player = {
			type = "group",
			order = 10,
			name = "Player",
			inline = true,
			args = {
				enable = {	
					type = "toggle",
					order = 1,
					width = "full",
					name = "Enable Player Repairs",
					desc = "Enable or disable player-funded automatic repairs.",
					descStyle = "inline",
					get = function( info ) return addon.db.profile.player.enabled end,
					set = function( info, value ) addon.db.profile.player.enabled = value end,
					disabled = _IsAddonDisabled,
				},
				reputation = {
					type = "select",
					order = 2,
					name = "Reputation",
					desc = "The minimum reputation required with the merchant to perform player-funded automatic repairs.",
					values = OPTIONS_REPUTATION_VALUES,
					get = function( info ) return addon.db.profile.player.reputation end,
					set = function( info, value ) addon.db.profile.player.reputation = value end,
					disabled = _IsPlayerRepairDisabled,
				},
				reservePrefix = { -- this is just to add whitespace in the interface options frame
					type = "description",
					order = 3,
					width = "full",
					name = " ",
				},
				reserve = {
					type = "input",
					order = 4,
					name = "Reserve",
					get = function( info ) return tostring( addon.db.profile.player.reserve ) end,
					set = function( info, value ) addon.db.profile.player.reserve = tonumber( value ) end,
					pattern = "^[%d]+$",
					usage = "This must be a numeric value.",
					disabled = _IsPlayerRepairDisabled,
				},
				reserveSuffix = {
					type = "description",
					order = 5,
					width = "full",
					name = "The minimum quantity of gold to reserve. Automatic repairs will not be performed if the player does not have at least this much gold.",
				},
			},
		},
		guild = {
			type = "group",
			order = 20,
			name = "Guild",
			inline = true,
			args = {
				enable = {
					type = "toggle",
					order = 1,
					width = "full",
					name = "Enable Guild Repairs",
					desc = "Enable or disable guild-funded automatic repairs.",
					descStyle = "inline",
					get = function( info ) return addon.db.profile.guild.enabled end,
					set = function( info, value ) addon.db.profile.guild.enabled = value end,
					disabled = _IsAddonDisabled,
				},
				reputation = {
					type = "select",
					order = 2,
					name = "Reputation",
					desc = "The minimum reputation required with the merchant to perform guild-funded automatic repairs.",
					values = OPTIONS_REPUTATION_VALUES,
					get = function( info ) return addon.db.profile.guild.reputation end,
					set = function( info, value ) addon.db.profile.guild.reputation = value end,
					disabled = _IsGuildRepairDisabled,
				},
				raidHeader = {
					type = "header",
					order = 10,
					name = "Raid",
				},
				guildRaid = {	
					type = "toggle",
					order = 11,
					width = "full",
					name = "Guild Raid",
					desc = "Enable or disable guild-funded automatic repairs while in a guild raid.",
					descStyle = "inline",
					get = function( info ) return addon.db.profile.guild.raid.enabled end,
					set = function( info, value ) addon.db.profile.guild.raid.enabled = value end,
					disabled = _IsGuildRepairDisabled,
				},
				anyRaid = {	
					type = "toggle",
					order = 12,
					width = "full",
					name = "Any Raid",
					desc = "Enable or disable guild-funded automatic repairs while in any raid.",
					descStyle = "inline",
					get = function( info ) return addon.db.profile.guild.raid.any end,
					set = function( info, value ) addon.db.profile.guild.raid.any = value end,
					disabled = _IsRaidRepairDisabled,
				},
				partyHeader = {
					type = "header",
					order = 20,
					name = "Party",
				},
				guildParty = {	
					type = "toggle",
					order = 21,
					width = "full",
					name = "Guild Party",
					desc = "Enable or disable guild-funded automatic repairs while in a guild party.",
					descStyle = "inline",
					get = function( info ) return addon.db.profile.guild.party.enabled end,
					set = function( info, value ) addon.db.profile.guild.party.enabled = value end,
					disabled = _IsGuildRepairDisabled,
				},
				anyParty = {	
					type = "toggle",
					order = 22,
					width = "full",
					name = "Any Party",
					desc = "Enable or disable guild-funded automatic repairs while in any party.",
					descStyle = "inline",
					get = function( info ) return addon.db.profile.guild.party.any end,
					set = function( info, value ) addon.db.profile.guild.party.any = value end,
					disabled = _IsPartyRepairDisabled
				},
				soloHeader = {
					type = "header",
					order = 30,
					name = "Solo",
				},
				guildSolo = {	
					type = "toggle",
					order = 31,
					width = "full",
					name = "No Group",
					desc = "Enable or disable guild-funded automatic repairs while not in a party or raid.",
					descStyle = "inline",
					get = function( info ) return addon.db.profile.guild.solo.enabled end,
					set = function( info, value ) addon.db.profile.guild.solo.enabled = value end,
					disabled = _IsGuildRepairDisabled,
				},
			},
		},
	},
}

local function _SetDefaultOptions( )
	addon.db:ResetProfile( )
	LibStub( "AceConfigRegistry-3.0" ):NotifyChange( addon.name )
end

local function CreateProfileOptionFrame( addon )
	addon.profiles = LibStub( "AceDBOptions-3.0" ):GetOptionsTable( addon.db )
	LibStub( "AceConfig-3.0" ):RegisterOptionsTable( addon.name .. "Profiles", addon.profiles )
	LibStub( "AceConfigDialog-3.0" ):AddToBlizOptions( addon.name .. "Profiles", "Profiles", addon.name )
end

function addon:OnInitialize( )
	self.db = LibStub( "AceDB-3.0" ):New( "hcRepairBotSavedVariables", DEFAULTS, true )
	
	LibStub( "AceConfig-3.0" ):RegisterOptionsTable( self.name, OPTIONS )
	local optionsFrame = LibStub( "AceConfigDialog-3.0" ):AddToBlizOptions( self.name, self.name )
	optionsFrame.default = _SetDefaultOptions
 
	CreateAboutOptionFrame( self.name )
	CreateProfileOptionFrame( self )

	self:SetEnabledState( not _IsAddonDisabled( ) )
end

function addon:OnEnable( )
	self.db.profile.enabled = true
	self:RegisterEvent( "MERCHANT_SHOW" )
end

function addon:OnDisable( )
	self.db.profile.enabled = false
	self:UnregisterEvent( "MERCHANT_SHOW" )
end

--
-- Tests whether or not the player meets the reputation criteria 
-- set for automatic repairs.
--
-- @param profile repair options
-- @return true if the player meets the reputation criteria, false 
-- otherwise.
--
local function _IsPlayerReputable( standing )
	return ( UnitReaction( "npc", "player" ) >= standing )
end

--
-- @return true if the player is in a raid group, false otherwise.
--
local function _IsPlayerInRaid( )
	return GetNumRaidMembers( ) > 0
end

--
-- @return the current raid size, as defined by the raid difficulty.
--
local function _GetRaidDifficultySize( )
	local difficulty = GetCurrentRaidDifficulty( )
	if ( difficulty == 1 or difficulty == 3 ) then
		return 10
	else
		return 25
	end
end

--
-- @return true if the player is in a raid group with eight or twenty 
-- guild members, for ten-and twenty-five-man raids, repsectively.
--
local function _IsPlayerInGuildRaid( )
	local threshold = _GetRaidDifficultySize( ) * 0.8
	local numGuildMembers = 0
	
	for raidId = 1, GetNumRaidMembers( ) do
		if UnitIsInMyGuild( "raid" .. raidId ) then
			numGuildMembers = numGuildMembers + 1
		end
	end
	
	return ( numGuildMembers >= threshold )
end

-- 
-- @return true if the player is in a party.
--
local function _IsPlayerInParty( )
	return GetNumPartyMembers( ) > 0
end

-- 
-- @return true if the player is in a party with four or more guild members.
--
local function _IsPlayerInGuildParty( )
	local threshold = 4
	for partyId = 1, numPartyMembers do
		numGuildMembers = numGuildMembers + 1
	end
	return ( numGuildMembers >= threshold )
end

--
-- Tests whether or not guild funds should be used to repair the 
-- player's gear.
--
-- @param profile repair options
-- @return true if guild funds should be used, false otherwise.
--
function addon:ShouldGuildRepair( cost )
	if not IsInGuild( ) then
		self:PrintV( "Player is not in a guild." )
		return false
	end
	
	if not CanGuildBankRepair( ) then
		self:PrintV( "Guild-funded repairs are not available to the player." )
		return false
	end
	
	if _IsGuildRepairDisabled( ) then
		self:PrintV( "Guild-funded automatic repairs are not enabled." )
		return false
	end
	
	if not _IsPlayerReputable( self.db.profile.guild.reputation ) then
		self:PrintV( "Guild-funded automatic repairs are not enabled for merchants with insufficient reputation." )
		return false
	end

	local budget = GetGuildBankWithdrawMoney( )
	if cost > budget then
		self:PrintV( "Guild repair funds are insufficient." )
		return false
	end
	
	if _IsPlayerInRaid( ) and self.db.profile.guild.raid.enabled then
		self:PrintV( "Guild-funded automatic repairs may be allowed for this raid group." )
		return _IsPlayerInGuildRaid( ) or profile.guild.raid.any
	end
	
	if _IsPlayerInParty( ) and self.db.profile.guild.party.enabled then
		self:PrintV( "Guild-funded automatic repairs may be allowed for this party." )
		return _IsPlayerInGuildParty( ) or profile.guild.party.any
	end
	
	if not self.db.profile.guild.solo.enabled then
		self:PrintV( "Guild-funded automatic repairs are not allowed while not in a group." )
		return false
	end
	
	return true
end

--
-- Tests whether or not player funds should be used to repair the 
-- player's gear.
-- 
-- @param profile repair options.
-- @return true if player funds should be used, false otherwise.
--
function addon:ShouldPlayerRepair( cost )
	if _IsPlayerRepairDisabled( ) then
		self:PrintV( "Player-funded automatic repairs are not enabled." )
		return false
	end
	
	if not _IsPlayerReputable( self.db.profile.player.reputation ) then
		self:PrintV( "Player-funded automatic repairs are not enabled for merchants with insufficient reputation." )
		return false
	end

	local budget = GetMoney( ) - ( self.db.profile.player.reserve * 10000 )
	if cost > budget then
		self:PrintV( "Player repair funds are insufficient." )
		return false
	end
	
	return true
end

function addon:MERCHANT_SHOW( )
	if not self.db.profile.enabled then
		-- No automatic repairs should be made.
		return
	end

	if not CanMerchantRepair( ) then
		-- No repairs are possible.
		return
	end

	local cost = GetRepairAllCost( )
	if ( cost <= 0 ) then
		-- No repairs are necessary.
		return
	end
	
	if self:ShouldGuildRepair( cost ) then
		RepairAllItems( 1 )
		self:Print( "Repaired all items for " .. GetCoinTextureString( cost ) .. " using guild funds." )
	elseif self:ShouldPlayerRepair( cost ) then
		RepairAllItems( )
		self:Print( "Repaired all items for " .. GetCoinTextureString( cost ) .. " using player funds." )
	end
end

--
-- Prints a given message, only if verbose output is enabled.
--
-- @param a message
--
function addon:PrintV( message )
	if self.db.profile.verbose then
		self:Print( message )
	end
end
